import os
from pathlib import Path
from typing import Any, Dict

from ramalama.common import container_manager, default_image
from ramalama.toml_parser import TOMLParser

DEFAULT_PORT_RANGE = (8080, 8090)


def get_store():
    if os.geteuid() == 0:
        return "/var/lib/ramalama"

    return os.path.expanduser("~/.local/share/ramalama")


def use_container():
    use_container = os.getenv("RAMALAMA_IN_CONTAINER")
    if use_container:
        return use_container.lower() == "true"

    conman = container_manager()
    return conman is not None


def load_config() -> Dict[str, Any]:
    """Load configuration from a list of paths, in priority order."""
    parser = TOMLParser()
    config_path = os.getenv("RAMALAMA_CONFIG")
    if config_path:
        return parser.parse_file(config_path)

    config = {}
    config_paths = [
        "/usr/share/ramalama/ramalama.conf",
        "/usr/local/share/ramalama/ramalama.conf",
        "/etc/ramalama/ramalama.conf",
    ]
    config_home = os.getenv("XDG_CONFIG_HOME", os.path.join("~", ".config"))
    config_paths.extend([os.path.expanduser(os.path.join(config_home, "ramalama", "ramalama.conf"))])

    # Load configuration from each path
    for path in config_paths:
        if os.path.exists(path):
            # Load the main config file
            config = parser.parse_file(path)
        if os.path.isdir(path + ".d"):
            # Load all .conf files in ramalama.conf.d directory
            for conf_file in sorted(Path(path + ".d").glob("*.conf")):
                config = parser.parse_file(conf_file)

    return config


def load_config_from_env(config: Dict[str, Any], env: Dict):
    """Load configuration from environment variables."""
    config['container'] = env.get('RAMALAMA_IN_CONTAINER', config.get('container', use_container()))
    config['engine'] = env.get('RAMALAMA_CONTAINER_ENGINE', config.get('engine', container_manager()))
    config['image'] = env.get('RAMALAMA_IMAGE', config.get('image', default_image()))
    config['store'] = env.get('RAMALAMA_STORE', config.get('store', get_store()))
    config['transport'] = env.get('RAMALAMA_TRANSPORT', config.get('transport', "ollama"))


def int_tuple_as_str(input: tuple) -> str:
    return '-'.join(map(str, input))


def load_config_defaults(config: Dict[str, Any]):
    """Set configuration defaults if these are not yet set."""
    config['carimage'] = config.get('carimage', "registry.access.redhat.com/ubi9-micro:latest")
    config['ctx_size'] = config.get('ctx_size', 2048)
    config['env'] = config.get('env', [])
    config['host'] = config.get('host', "0.0.0.0")
    config['keep_groups'] = config.get('keep_groups', False)
    config['ngl'] = config.get('ngl', -1)
    config['threads'] = config.get('threads', -1)
    config['nocontainer'] = config.get('nocontainer', False)
    config['port'] = config.get('port', int_tuple_as_str(DEFAULT_PORT_RANGE))
    config['pull'] = config.get('pull', "newer")
    config['runtime'] = config.get('runtime', 'llama.cpp')
    config['temp'] = config.get('temp', "0.8")
    config['use_model_store'] = config.get('use_model_store', False)
    config['images'] = config.get(
        'images',
        {
            "ASAHI_VISIBLE_DEVICES": "quay.io/ramalama/asahi",
            "ASCEND_VISIBLE_DEVICES": "quay.io/ramalama/cann",
            "CUDA_VISIBLE_DEVICES": "quay.io/ramalama/cuda",
            "HIP_VISIBLE_DEVICES": "quay.io/ramalama/rocm-fedora",
            "INTEL_VISIBLE_DEVICES": "quay.io/ramalama/intel-gpu",
        },
    )


def load_and_merge_config() -> Dict[str, Any]:
    """Load configuration from files, merge with environment variables and set defaults for options not set."""
    config = load_config()

    ramalama_config = config.setdefault('ramalama', {})
    load_config_from_env(ramalama_config, os.environ)
    load_config_defaults(ramalama_config)

    return ramalama_config


CONFIG = load_and_merge_config()



import logging
import os
import shutil
import urllib
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import ramalama.go2jinja as go2jinja
import ramalama.oci
from ramalama.common import download_file, generate_sha256, verify_checksum
from ramalama.gguf_parser import GGUFInfoParser, GGUFModelInfo

LOGGER = logging.getLogger(__name__)


def sanitize_hash(filename: str) -> str:
    return filename.replace(":", "-")


class SnapshotFileType(IntEnum):
    Model = 1
    ChatTemplate = 2
    Other = 3


class SnapshotFile:

    def __init__(
        self,
        url: str,
        header: Dict,
        hash: str,
        name: str,
        type: SnapshotFileType,
        should_show_progress: bool = False,
        should_verify_checksum: bool = False,
        required: bool = True,
    ):
        self.url: str = url
        self.header: Dict = header
        self.hash: str = hash
        self.name: str = name
        self.type: SnapshotFileType = type
        self.should_show_progress: bool = should_show_progress
        self.should_verify_checksum: bool = should_verify_checksum
        self.required: bool = required

    def download(self, blob_file_path: str, snapshot_dir: str) -> str:
        download_file(
            url=self.url,
            headers=self.header,
            dest_path=blob_file_path,
            show_progress=self.should_show_progress,
        )
        return os.path.relpath(blob_file_path, start=snapshot_dir)


class LocalSnapshotFile(SnapshotFile):

    def __init__(
        self,
        content: str,
        name: str,
        type: SnapshotFileType,
        should_show_progress: bool = False,
        should_verify_checksum: bool = False,
        required: bool = True,
    ):
        super().__init__(
            "", "", generate_sha256(content), name, type, should_show_progress, should_verify_checksum, required
        )
        self.content = content

    def download(self, blob_file_path, snapshot_dir):
        with open(blob_file_path, "w") as file:
            file.write(self.content)
            file.flush()
        return os.path.relpath(blob_file_path, start=snapshot_dir)


def validate_snapshot_files(snapshot_files: list[SnapshotFile]):
    model_files = []
    chat_template_files = []
    for file in snapshot_files:
        if file.type == SnapshotFileType.Model:
            model_files.append(file)
        if file.type == SnapshotFileType.ChatTemplate:
            chat_template_files.append(file)

    if len(model_files) > 1:
        raise Exception(f"Only one model file supported, got {len(model_files)}: {model_files}")
    if len(chat_template_files) > 1:
        raise Exception(f"Only one chat template file supported, got {len(chat_template_files)}: {chat_template_files}")


class RefFile:

    SEP = "---"
    MODEL_SUFFIX = "model"
    CHAT_TEMPLATE_SUFFIX = "chat"

    def __init__(self):
        self.hash: str = ""
        self.filenames: list[str] = []
        self.model_name: str = ""
        self.chat_template_name: str = ""
        self._path: str = ""

    @property
    def path(self) -> str:
        return self._path

    def from_path(path: str) -> "RefFile":
        ref_file = RefFile()
        ref_file._path = path
        with open(path, "r") as file:
            ref_file.hash = file.readline().strip()
            filename = file.readline().strip()
            while filename != "":
                parts = filename.split(RefFile.SEP)
                if len(parts) != 2:
                    ref_file.filenames.append(filename)
                    filename = file.readline().strip()
                    continue

                ref_file.filenames.append(parts[0])
                if parts[1] == RefFile.MODEL_SUFFIX:
                    ref_file.model_name = parts[0]
                if parts[1] == RefFile.CHAT_TEMPLATE_SUFFIX:
                    ref_file.chat_template_name = parts[0]

                filename = file.readline().strip()
        return ref_file

    def serialize(self) -> str:
        lines = [self.hash]
        for filename in self.filenames:
            line = f"{filename}{RefFile.SEP}"
            if filename == self.model_name:
                line = line + RefFile.MODEL_SUFFIX
            if filename == self.chat_template_name:
                line = line + RefFile.CHAT_TEMPLATE_SUFFIX
            lines.append(line)
        return "\n".join(lines)


@dataclass
class ModelFile:
    name: str
    modified: float
    size: int


DIRECTORY_NAME_BLOBS = "blobs"
DIRECTORY_NAME_REFS = "refs"
DIRECTORY_NAME_SNAPSHOTS = "snapshots"


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class GlobalModelStore:

    def __init__(
        self,
        base_path: Path,
    ):
        self._store_base_path = os.path.join(base_path, "store")

    @property
    def path(self) -> str:
        return self._store_base_path

    def list_models(self, engine: str, debug: bool) -> Dict[str, List[ModelFile]]:
        models: Dict[str, List[ModelFile]] = {}

        for root, subdirs, _ in os.walk(self.path):
            if DIRECTORY_NAME_REFS in subdirs:
                ref_dir = os.path.join(root, DIRECTORY_NAME_REFS)
                for ref_file_name in os.listdir(ref_dir):
                    ref_file: RefFile = RefFile.from_path(os.path.join(ref_dir, ref_file_name))
                    model_path = root.replace(self.path, "").replace(os.sep, "", 1)
                    model_name = f"{model_path}:{ref_file_name}"

                    models[model_name] = []
                    for snapshot_file in ref_file.filenames:
                        snapshot_file_path = os.path.join(root, DIRECTORY_NAME_SNAPSHOTS, ref_file.hash, snapshot_file)
                        last_modified = os.path.getmtime(snapshot_file_path)
                        file_size = os.path.getsize(snapshot_file_path)
                        models[model_name].append(ModelFile(snapshot_file, last_modified, file_size))

        oci_models = ramalama.oci.list_models(
            dotdict(
                {
                    "engine": engine,
                    "debug": debug,
                }
            )
        )
        for oci_model in oci_models:
            name, modified, size = (oci_model["name"], oci_model["modified"], oci_model["size"])
            # ramalama.oci.list_models provides modified as timestamp string, convert it to unix timestamp
            modified_unix = datetime.fromisoformat(modified).timestamp()
            models[name] = [ModelFile(name, modified_unix, size)]

        return models

    # TODO:
    # iterating over all symlinks in snapshot dir, check valid
    def verify_snapshot(self):
        pass

    # TODO:
    # iterating over models and check
    #    1. for broken symlinks in snapshot dirs -> delete and update refs
    #    2. for blobs not reached by ref->snapshot chain -> delete
    #    3. for empty folders -> delete
    def cleanup(self):
        pass


class ModelStore:

    def __init__(
        self,
        store: GlobalModelStore,
        model_name: str,
        model_type: str,
        model_organization: str,
    ):
        self._store = store
        self._model_name = model_name
        self._model_type = model_type
        self._model_organization = model_organization

    @property
    def base_path(self) -> str:
        return self._store.path

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def model_organization(self) -> str:
        return self._model_organization if self._model_organization != "" else self._model_name

    @property
    def model_type(self) -> str:
        return self._model_type

    @property
    def model_base_directory(self) -> str:
        return os.path.join(self.base_path, self.model_type, self.model_organization)

    @property
    def blobs_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_BLOBS)

    @property
    def refs_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_REFS)

    @property
    def snapshots_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_SNAPSHOTS)

    def tag_exists(self, model_tag: str) -> bool:
        return os.path.exists(self.get_ref_file_path(model_tag))

    def file_exists(self, file_path: str) -> bool:
        return os.path.exists(file_path)

    def get_ref_file_path(self, model_tag: str) -> str:
        return os.path.join(self.refs_directory, model_tag)

    def get_ref_file(self, model_tag: str) -> Optional[RefFile]:
        if not self.tag_exists(self.get_ref_file_path(model_tag)):
            return None

        return RefFile.from_path(self.get_ref_file_path(model_tag))

    def update_ref_file(
        self, model_tag: str, snapshot_hash: str = "", snapshot_files: list[SnapshotFile] = []
    ) -> Optional[RefFile]:
        ref_file_path = self.get_ref_file_path(model_tag)
        if not os.path.exists(ref_file_path):
            return None

        ref_file: RefFile = RefFile.from_path(self.get_ref_file_path(model_tag))
        if snapshot_hash != "":
            ref_file.hash = snapshot_hash
        if snapshot_files != []:
            ref_file.filenames = [file.name for file in snapshot_files]

        with open(ref_file_path, "w") as file:
            file.write(ref_file.serialize())
            file.flush()

        return ref_file

    def get_snapshot_hash(self, model_tag: str) -> str:
        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            return ""
        return sanitize_hash(ref_file.hash)

    def get_snapshot_directory_from_tag(self, model_tag: str) -> str:
        return os.path.join(self.snapshots_directory, self.get_snapshot_hash(model_tag))

    def get_snapshot_directory(self, hash: str) -> str:
        return os.path.join(self.snapshots_directory, hash)

    def get_snapshot_file_path(self, tag_hash: str, filename: str) -> str:
        return os.path.join(self.snapshots_directory, sanitize_hash(tag_hash), filename)

    def get_blob_file_path(self, file_hash: str) -> str:
        return os.path.join(self.blobs_directory, sanitize_hash(file_hash))

    def get_partial_blob_file_path(self, file_hash: str) -> str:
        return self.get_blob_file_path(file_hash) + ".partial"

    def ensure_directory_setup(self) -> None:
        os.makedirs(self.blobs_directory, exist_ok=True)
        os.makedirs(self.refs_directory, exist_ok=True)
        os.makedirs(self.snapshots_directory, exist_ok=True)

    def directory_setup_exists(self) -> bool:
        return (
            os.path.exists(self.blobs_directory)
            and os.path.exists(self.refs_directory)
            and os.path.exists(self.snapshots_directory)
        )

    def get_cached_files(self, model_tag: str) -> Tuple[str, list[str], bool]:
        cached_files = []

        ref_file_path = self.get_ref_file_path(model_tag)
        if not self.file_exists(ref_file_path):
            return ("", cached_files, False)

        ref_file: RefFile = RefFile.from_path(ref_file_path)
        for file in ref_file.filenames:
            path = self.get_snapshot_file_path(ref_file.hash, file)
            if os.path.exists(path):
                cached_files.append(file)

        return (ref_file.hash, cached_files, len(cached_files) == len(ref_file.filenames))

    def _prepare_new_snapshot(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
        validate_snapshot_files(snapshot_files)
        self.ensure_directory_setup()

        ref_file_path = self.get_ref_file_path(model_tag)
        if not self.file_exists(ref_file_path):
            ref_file = RefFile()
            ref_file.hash = snapshot_hash
            ref_file.filenames = [file.name for file in snapshot_files]
            for file in snapshot_files:
                if file.type == SnapshotFileType.Model:
                    ref_file.model_name = file.name
                if file.type == SnapshotFileType.ChatTemplate:
                    ref_file.chat_template_name = file.name
            with open(ref_file_path, "w") as file:
                file.write(ref_file.serialize())
                file.flush()

        snapshot_directory = self.get_snapshot_directory(snapshot_hash)
        os.makedirs(snapshot_directory, exist_ok=True)

    def _download_snapshot_files(self, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
        for file in snapshot_files:
            dest_path = self.get_blob_file_path(file.hash)
            blob_relative_path = ""
            try:
                blob_relative_path = file.download(dest_path, self.get_snapshot_directory(snapshot_hash))
            except urllib.error.HTTPError:
                if file.required:
                    raise
                continue

            if file.should_verify_checksum:
                if not verify_checksum(dest_path):
                    LOGGER.info(f"Checksum mismatch for blob {dest_path}, retrying download...")
                    os.remove(dest_path)
                    file.download(dest_path, self.get_snapshot_directory(snapshot_hash))
                    if not verify_checksum(dest_path):
                        raise ValueError(f"Checksum verification failed for blob {dest_path}")

            os.symlink(blob_relative_path, self.get_snapshot_file_path(snapshot_hash, file.name))

    def _ensure_chat_template(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
        model_file: SnapshotFile = None
        for file in snapshot_files:
            # Give preference to a chat template that has been specified in the file list
            if file.type == SnapshotFileType.ChatTemplate:
                chat_template_file_path = self.get_blob_file_path(file.hash)
                chat_template = ""
                with open(chat_template_file_path, "r") as file:
                    chat_template = file.read()

                if not go2jinja.is_go_template(chat_template):
                    return

                jinja_template = go2jinja.go_to_jinja(chat_template)
                files = [LocalSnapshotFile(jinja_template, "chat_template_converted", SnapshotFileType.ChatTemplate)]
                self.update_snapshot(model_tag, snapshot_hash, files)
                return
            if file.type == SnapshotFileType.Model:
                model_file = file

        model_file_path = self.get_blob_file_path(model_file.hash)
        if not GGUFInfoParser.is_model_gguf(model_file_path):
            return

        # Parse model, first and second parameter are irrelevant here
        info: GGUFModelInfo = GGUFInfoParser.parse("model", "registry", model_file_path)
        tmpl = info.get_chat_template()
        if tmpl == "":
            return

        files = [LocalSnapshotFile(tmpl, "chat_template", SnapshotFileType.ChatTemplate)]
        if go2jinja.is_go_template(tmpl):
            jinja_template = go2jinja.go_to_jinja(tmpl)
            files.append(LocalSnapshotFile(jinja_template, "chat_template_converted", SnapshotFileType.ChatTemplate))

        self.update_snapshot(model_tag, snapshot_hash, files)

    def new_snapshot(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
        snapshot_hash = sanitize_hash(snapshot_hash)
        self._prepare_new_snapshot(model_tag, snapshot_hash, snapshot_files)
        self._download_snapshot_files(snapshot_hash, snapshot_files)
        self._ensure_chat_template(model_tag, snapshot_hash, snapshot_files)

    def update_snapshot(self, model_tag: str, snapshot_hash: str, new_snapshot_files: list[SnapshotFile]) -> bool:
        validate_snapshot_files(new_snapshot_files)
        snapshot_hash = sanitize_hash(snapshot_hash)

        if not self.directory_setup_exists():
            return False

        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            return False

        ref_file.filenames = ref_file.filenames + [file.name for file in new_snapshot_files]
        # update model and chat template name
        for file in new_snapshot_files:
            if file.type == SnapshotFileType.Model:
                ref_file.model_name = file.name
            if file.type == SnapshotFileType.ChatTemplate:
                ref_file.chat_template_name = file.name

        with open(ref_file.path, "w") as file:
            file.write(ref_file.serialize())
            file.flush()

        self._download_snapshot_files(snapshot_hash, new_snapshot_files)
        return True

    def _remove_blob_file(self, snapshot_file_path: str):
        blob_path = Path(snapshot_file_path).resolve()
        try:
            if os.path.exists(blob_path) and self.base_path in blob_path.parents:
                os.remove(blob_path)
                LOGGER.debug(f"Removed blob for '{snapshot_file_path}'")
        except Exception as ex:
            LOGGER.error(f"Failed to remove blob file '{blob_path}': {ex}")

    def remove_snapshot(self, model_tag: str):
        ref_file = self.get_ref_file(model_tag)

        # Remove all blobs first
        if ref_file is not None:
            for file in ref_file.filenames:
                self._remove_blob_file(self.get_snapshot_file_path(ref_file.hash, file))
                self._remove_blob_file(self.get_partial_blob_file_path(ref_file.hash))

        # Remove snapshot directory
        snapshot_directory = self.get_snapshot_directory_from_tag(model_tag)
        try:
            shutil.rmtree(snapshot_directory, ignore_errors=False)
        except Exception as ex:
            LOGGER.error(f"Failed to remove snapshot directory '{snapshot_directory}': {ex}")
            # only continue to remove the ref file when blobs and snapshot directory have been removed
            return

        # Remove ref file
        ref_file_path = self.get_ref_file_path(model_tag)
        try:
            os.remove(ref_file_path)
        except Exception as ex:
            LOGGER.error(f"Failed to remove ref file '{ref_file_path}': {ex}")



import os

from ramalama.common import MNT_CHAT_TEMPLATE_FILE, MNT_DIR, MNT_FILE, RAG_DIR, get_accel_env_vars


class Quadlet:
    def __init__(self, model, chat_template, image, args, exec_args):
        self.ai_image = model
        if hasattr(args, "MODEL"):
            self.ai_image = args.MODEL
        self.ai_image = self.ai_image.removeprefix("oci://")
        if args.name:
            self.name = args.name
        else:
            self.name = os.path.basename(self.ai_image)

        self.model = model.removeprefix("oci://")
        self.args = args
        self.exec_args = exec_args
        self.image = image
        self.chat_template = chat_template
        self.rag = ""
        self.rag_name = ""
        if args.rag:
            self.rag = args.rag.removeprefix("oci://")
            self.rag_name = os.path.basename(self.rag) + "-rag"

    def kube(self):
        outfile = self.name + ".kube"
        print(f"Generating quadlet file: {outfile}")
        with open(outfile, 'w') as c:
            c.write(
                f"""\
[Unit]
Description=RamaLama {self.model} Kubernetes YAML - AI Model Service
After=local-fs.target

[Kube]
Yaml={self.name}.yaml

[Install]
# Start by default on boot
WantedBy=multi-user.target default.target
"""
            )

    def generate(self):
        chat_template_volume = self._gen_chat_template_volume()
        env_var_string = self._gen_env()
        model_volume = self._gen_model_volume()
        name_string = self._gen_name()
        port_string = self._gen_port()
        rag_volume = self._gen_rag_volume()

        outfile = self.name + ".container"
        print(f"Generating quadlet file: {outfile}")
        with open(outfile, 'w') as c:
            c.write(
                f"""\
[Unit]
Description=RamaLama {self.model} AI Model Service
After=local-fs.target

[Container]
AddDevice=-/dev/accel
AddDevice=-/dev/dri
AddDevice=-/dev/kfd\
{env_var_string}
Exec={" ".join(self.exec_args)}
Image={self.image}\
{model_volume}\
{rag_volume}\
{chat_template_volume}\
{name_string}\
{port_string}

[Install]
# Start by default on boot
WantedBy=multi-user.target default.target
"""
            )

    def _gen_chat_template_volume(self):
        if os.path.exists(self.chat_template):
            return f"\nMount=type=bind,src={self.chat_template},target={MNT_CHAT_TEMPLATE_FILE},ro,Z"
        return ""

    def _gen_env(self):
        env_var_string = ""
        for k, v in get_accel_env_vars().items():
            env_var_string += f"\nEnvironment={k}={v}"
        for e in self.args.env:
            env_var_string += f"\nEnvironment={e}"
        return env_var_string

    def _gen_image(self, name, image):
        outfile = name + ".image"
        print(f"Generating quadlet file: {outfile} ")
        with open(outfile, 'w') as c:
            c.write(
                f"""\
[Image]
Image={image}
"""
            )

    def _gen_name(self):
        name_string = ""
        if hasattr(self.args, "name") and self.args.name:
            name_string = f"\nContainerName={self.args.name}"
        return name_string

    def _gen_model_volume(self):
        if os.path.exists(self.model):
            return f"\nMount=type=bind,src={self.model},target={MNT_FILE},ro,Z"

        outfile = self.name + ".volume"

        print(f"Generating quadlet file: {outfile} ")
        with open(outfile, 'w') as c:
            c.write(
                f"""\
[Volume]
Driver=image
Image={self.name}.image
"""
            )
        self._gen_image(self.name, self.ai_image)
        return f"\nMount=type=image,source={self.ai_image},destination={MNT_DIR},subpath=/models,readwrite=false"

    def _gen_port(self):
        port_string = ""
        if hasattr(self.args, "port"):
            port_string = f"\nPublishPort={self.args.port}"
        return port_string

    def _gen_rag_volume(self):
        rag_volume = ""
        if not hasattr(self.args, "rag") or not self.rag:
            return rag_volume

        outfile = self.rag_name + ".volume"

        print(f"Generating quadlet file: {outfile} ")
        with open(outfile, 'w') as c:
            c.write(
                f"""\
[Volume]
Driver=image
Image={self.rag_name}.image
"""
            )
        self._gen_image(self.rag_name, self.rag)
        return f"\nMount=type=image,source={self.rag},destination={RAG_DIR},readwrite=false"



import glob
import os
import platform
import random
import socket
import sys

import ramalama
from ramalama.common import (
    MNT_CHAT_TEMPLATE_FILE,
    MNT_DIR,
    MNT_FILE,
    accel_image,
    check_nvidia,
    exec_cmd,
    genname,
    get_accel_env_vars,
    run_cmd,
    set_accel_env_vars,
)
from ramalama.config import CONFIG, DEFAULT_PORT_RANGE, int_tuple_as_str
from ramalama.console import EMOJI
from ramalama.gguf_parser import GGUFInfoParser
from ramalama.kube import Kube
from ramalama.model_inspect import GGUFModelInfo, ModelInfoBase
from ramalama.model_store import ModelStore
from ramalama.quadlet import Quadlet
from ramalama.version import version

MODEL_TYPES = ["file", "https", "http", "oci", "huggingface", "hf", "ollama"]


file_not_found = """\
RamaLama requires the "%(cmd)s" command to be installed on the host when running with --nocontainer.
RamaLama is designed to run AI Models inside of containers, where "%(cmd)s" is already installed.
Either install a package containing the "%(cmd)s" command or run the workload inside of a container.
%(error)s"""

file_not_found_in_container = """\
RamaLama requires the "%(cmd)s" command to be installed inside of the container.
RamaLama requires the server application be installed in the container images.
Either install a package containing the "%(cmd)s" command in the container or run
with the default RamaLama
$(error)s"""


class ModelBase:

    def __not_implemented_error(self, param):
        return NotImplementedError(f"ramalama {param} for '{type(self).__name__}' not implemented")

    def login(self, args):
        raise self.__not_implemented_error("login")

    def logout(self, args):
        raise self.__not_implemented_error("logout")

    def pull(self, args):
        raise self.__not_implemented_error("pull")

    def push(self, source, args):
        raise self.__not_implemented_error("push")

    def remove(self, args):
        raise self.__not_implemented_error("rm")

    def bench(self, args):
        raise self.__not_implemented_error("bench")

    def run(self, args):
        raise self.__not_implemented_error("run")

    def perplexity(self, args):
        raise self.__not_implemented_error("perplexity")

    def serve(self, args):
        raise self.__not_implemented_error("serve")

    def exists(self, args):
        raise self.__not_implemented_error("exists")

    def inspect(self, args):
        raise self.__not_implemented_error("inspect")


class Model(ModelBase):
    """Model super class"""

    model = ""
    type = "Model"

    def __init__(self, model):
        self.model = model

        split = self.model.rsplit("/", 1)
        self.directory = split[0] if len(split) > 1 else ""
        self.filename = split[1] if len(split) > 1 else split[0]

        self._model_name: str
        self._model_tag: str
        self._model_organization: str
        self._model_type: str
        self._model_name, self._model_tag, self._model_organization = self.extract_model_identifiers()
        self._model_type = type(self).__name__.lower()

        self.store: ModelStore = None

    def extract_model_identifiers(self):
        model_name = self.model
        model_tag = "latest"
        model_organization = ""

        # extract model tag from name if exists
        if ":" in model_name:
            model_name, model_tag = model_name.split(":", 1)

        # extract model organization from name if exists and update name
        split = model_name.rsplit("/", 1)
        model_organization = split[0].removeprefix("/") if len(split) > 1 else ""
        model_name = split[1] if len(split) > 1 else split[0]

        return model_name, model_tag, model_organization

    @property
    def name(self) -> str:
        return self._model_name

    @property
    def tag(self) -> str:
        return self._model_tag

    @property
    def organization(self) -> str:
        return self._model_organization

    @property
    def model_type(self) -> str:
        return self._model_type

    def is_symlink_to(self, file_path, target_path):
        if os.path.islink(file_path):
            symlink_target = os.readlink(file_path)
            abs_symlink_target = os.path.abspath(os.path.join(os.path.dirname(file_path), symlink_target))
            abs_target_path = os.path.abspath(target_path)
            return abs_symlink_target == abs_target_path

        return False

    def garbage_collection(self, args):
        for repo in MODEL_TYPES:
            repo_dir = f"{args.store}/repos/{repo}"
            model_dir = f"{args.store}/models/{repo}"
            for root, dirs, files in os.walk(repo_dir):
                file_has_a_symlink = False
                for file in files:
                    file_path = os.path.join(root, file)
                    if file.startswith("sha256:") or file.endswith(".gguf"):
                        file_path = os.path.join(root, file)
                        for model_root, model_dirs, model_files in os.walk(model_dir):
                            for model_file in model_files:
                                if self.is_symlink_to(os.path.join(root, model_root, model_file), file_path):
                                    file_has_a_symlink = True

                        if not file_has_a_symlink:
                            os.remove(file_path)
                            file_path = os.path.basename(file_path)
                            print(f"Deleted: {file_path}")

    def remove(self, args):
        if self.store is not None:
            _, tag, _ = self.extract_model_identifiers()
            self.store.remove_snapshot(tag)
            return

        model_path = self.model_path(args)
        try:
            os.remove(model_path)
            print(f"Untagged: {self.model}")
        except OSError as e:
            if not args.ignore:
                raise KeyError(f"removing {self.model}: {e}")
        self.garbage_collection(args)

    def get_container_name(self, args):
        if hasattr(args, "name") and args.name:
            return args.name

        return genname()

    def get_base_conman_args(self, args, name):
        return [
            args.engine,
            "run",
            "--rm",
            "-i",
            "--label",
            "ai.ramalama",
            "--name",
            name,
            "--env=HOME=/tmp",
            "--init",
        ]

    def add_privileged_options(self, conman_args, args):
        if args.privileged:
            conman_args += ["--privileged"]
        else:
            conman_args += [
                "--security-opt=label=disable",
                "--cap-drop=all",
                "--security-opt=no-new-privileges",
            ]

        return conman_args

    def add_container_labels(self, conman_args, args):
        if hasattr(args, "MODEL"):
            conman_args += ["--label", f"ai.ramalama.model={args.MODEL}"]

        if hasattr(args, "engine"):
            conman_args += ["--label", f"ai.ramalama.engine={args.engine}"]

        if hasattr(args, "runtime"):
            conman_args += ["--label", f"ai.ramalama.runtime={args.runtime}"]

        if hasattr(args, "port"):
            conman_args += ["--label", f"ai.ramalama.port={args.port}"]

        if hasattr(args, "subcommand"):
            conman_args += ["--label", f"ai.ramalama.command={args.subcommand}"]

        return conman_args

    def add_subcommand_env(self, conman_args, args):
        if EMOJI and hasattr(args, "subcommand") and args.subcommand == "run":
            if os.path.basename(args.engine) == "podman":
                conman_args += ["--env", "LLAMA_PROMPT_PREFIX=🦭 > "]
            elif os.path.basename(args.engine) == "docker":
                conman_args += ["--env", "LLAMA_PROMPT_PREFIX=🐋 > "]

        return conman_args

    def handle_podman_specifics(self, conman_args, args):
        if os.path.basename(args.engine) == "podman" and args.podman_keep_groups:
            conman_args += ["--group-add", "keep-groups"]

        return conman_args

    def handle_oci_pull(self, conman_args, args):
        self.image = accel_image(CONFIG, args)
        if not args.dryrun and os.path.basename(args.engine) == "docker" and args.pull == "newer":
            try:
                if not args.quiet:
                    print(f"Checking for newer image {self.image}")
                run_cmd([args.engine, "pull", "-q", args.image], ignore_all=True)
            except Exception:  # Ignore errors, the run command will handle it.
                pass
        else:
            conman_args += [f"--pull={args.pull}"]

        return conman_args

    def add_env_option(self, conman_args, args):
        for env in args.env:
            conman_args += ["--env", env]

        return conman_args

    def add_tty_option(self, conman_args):
        if sys.stdout.isatty() or sys.stdin.isatty():
            conman_args += ["-t"]

        return conman_args

    def add_detach_option(self, conman_args, args):
        if hasattr(args, "detach") and args.detach is True:
            conman_args += ["-d"]

        return conman_args

    def add_port_option(self, conman_args, args):
        if hasattr(args, "port"):
            conman_args += ["-p", f"{args.port}:{args.port}"]

        return conman_args

    def add_device_options(self, conman_args, args):
        if args.device:
            for device_arg in args.device:
                conman_args += ["--device", device_arg]

        if ramalama.common.podman_machine_accel:
            conman_args += ["--device", "/dev/dri"]

        for path in ["/dev/dri", "/dev/kfd", "/dev/accel", "/dev/davinci*", "/dev/devmm_svm", "/dev/hisi_hdc"]:
            for dev in glob.glob(path):
                conman_args += ["--device", dev]

        for k, v in get_accel_env_vars().items():
            # Special case for Cuda
            if k == "CUDA_VISIBLE_DEVICES":
                if os.path.basename(args.engine) == "docker":
                    conman_args += ["--gpus", "all"]
                else:
                    # newer Podman versions support --gpus=all, but < 5.0 do not
                    conman_args += ["--device", "nvidia.com/gpu=all"]

            conman_args += ["-e", f"{k}={v}"]

        return conman_args

    def add_network_option(self, conman_args, args):
        if args.network:
            conman_args += ["--network", args.network]

        return conman_args

    def add_oci_runtime(self, conman_args, args):
        if args.oci_runtime:
            conman_args += ["--runtime", args.oci_runtime]
        elif check_nvidia() == "cuda":
            if os.path.basename(args.engine) == "docker":
                conman_args += ["--runtime", "nvidia"]
            else:
                conman_args += ["--runtime", "/usr/bin/nvidia-container-runtime"]

        return conman_args

    def add_rag(self, exec_args, args):
        if not hasattr(args, "rag") or not args.rag:
            return exec_args

        if os.path.exists(args.rag):
            rag = os.path.realpath(args.rag)
            # Added temp read write because vector database requires write access even if nothing is written
            exec_args.append(f"--mount=type=bind,source={rag},destination=/rag/vector.db,rw=true")
        else:
            exec_args.append(f"--mount=type=image,source={args.rag},destination=/rag,rw=true")

        return exec_args

    def setup_container(self, args):
        if not args.engine:
            return []

        name = self.get_container_name(args)
        conman_args = self.get_base_conman_args(args, name)
        conman_args = self.add_oci_runtime(conman_args, args)
        conman_args = self.add_privileged_options(conman_args, args)
        conman_args = self.add_container_labels(conman_args, args)
        conman_args = self.add_subcommand_env(conman_args, args)
        conman_args = self.handle_podman_specifics(conman_args, args)
        conman_args = self.handle_oci_pull(conman_args, args)
        conman_args = self.add_tty_option(conman_args)
        conman_args = self.add_env_option(conman_args, args)
        conman_args = self.add_detach_option(conman_args, args)
        conman_args = self.add_port_option(conman_args, args)
        conman_args = self.add_device_options(conman_args, args)
        conman_args = self.add_network_option(conman_args, args)
        conman_args = self.add_rag(conman_args, args)

        return conman_args

    def gpu_args(self, args, runner=False):
        gpu_args = []
        machine = platform.machine()
        if (
            os.getenv("HIP_VISIBLE_DEVICES")
            or os.getenv("ASAHI_VISIBLE_DEVICES")
            or os.getenv("CUDA_VISIBLE_DEVICES")
            or os.getenv("INTEL_VISIBLE_DEVICES")
            or os.getenv("ASCEND_VISIBLE_DEVICES")
            or (
                # linux and macOS report aarch64 (linux), arm64 (macOS)
                ramalama.common.podman_machine_accel
                or (machine == "aarch64" and os.path.exists("/dev/dri"))
            )
        ):
            if args.ngl < 0:
                args.ngl = 999

            if runner:
                gpu_args += ["--ngl"]  # double dash
            else:
                gpu_args += ["-ngl"]  # single dash

            gpu_args += [f'{args.ngl}']

        # for some reason the --threads option is blowing up on Docker,
        # with option not being supported by llama-run.
        # This could be something being masked in a Docker container but not
        # in a Podman container.
        if args.threads != -1 and args.engine and os.path.basename(args.engine) != "docker":
            gpu_args += ["--threads", f"{args.threads}"]

        return gpu_args

    def exec_model_in_container(self, model_path, cmd_args, args):
        if not args.container:
            return False
        conman_args = self.setup_container(args)
        if len(conman_args) == 0:
            return False

        if model_path and os.path.exists(model_path):
            conman_args += [f"--mount=type=bind,src={model_path},destination={MNT_FILE},ro"]
        else:
            conman_args += [f"--mount=type=image,src={self.model},destination={MNT_DIR},subpath=/models"]

        # If a chat template is available, mount it as well
        if self.store is not None:
            ref_file = self.store.get_ref_file(self.tag)
            if ref_file.chat_template_name != "":
                chat_template_path = self.store.get_snapshot_file_path(ref_file.hash, ref_file.chat_template_name)
                conman_args += [f"--mount=type=bind,src={chat_template_path},destination={MNT_CHAT_TEMPLATE_FILE},ro"]

        # Make sure Image precedes cmd_args.
        conman_args += [accel_image(CONFIG, args)] + cmd_args

        if args.dryrun:
            dry_run(conman_args)
            return True

        exec_cmd(conman_args, debug=args.debug)
        return True

    def bench(self, args):
        model_path = self.get_model_path(args)
        exec_args = self.build_exec_args_bench(args, model_path)
        self.validate_args(args)
        self.execute_model(model_path, exec_args, args)

    def run(self, args):
        self.validate_args(args)
        prompt = self.build_prompt(args)
        model_path = self.get_model_path(args)
        exec_args = self.build_exec_args_run(args, model_path, prompt)
        if args.keepalive:
            exec_args = ["timeout", args.keepalive] + exec_args
        self.execute_model(model_path, exec_args, args)

    def perplexity(self, args):
        self.validate_args(args)
        model_path = self.get_model_path(args)
        exec_args = self.build_exec_args_perplexity(args, model_path)
        self.execute_model(model_path, exec_args, args)

    def build_exec_args_perplexity(self, args, model_path):
        exec_model_path = MNT_FILE if args.container else model_path
        exec_args = ["llama-perplexity"]

        set_accel_env_vars()
        gpu_args = self.gpu_args(args=args)
        if gpu_args is not None:
            exec_args.extend(gpu_args)

        exec_args += ["-m", exec_model_path]

        return exec_args

    def build_prompt(self, args):
        prompt = ""
        if args.ARGS:
            prompt = " ".join(args.ARGS)

        if not sys.stdin.isatty():
            inp = sys.stdin.read()
            prompt = inp + "\n\n" + prompt

        return prompt

    def model_path(self, args):
        if self.store is not None:
            _, tag, _ = self.extract_model_identifiers()
            if self.store.tag_exists(tag):
                fhash, _, _ = self.store.get_cached_files(tag)
                return self.store.get_snapshot_file_path(fhash, self.store.model_name)
            return ""

        return os.path.join(args.store, "models", self.type, self.directory, self.filename)

    def exists(self, args):
        model_path = self.model_path(args)
        if not os.path.exists(model_path):
            return None

        return model_path

    def get_model_path(self, args):
        model_path = self.exists(args)
        if model_path:
            return model_path

        if args.dryrun:
            return "/path/to/model"

        model_path = self.pull(args)

        return model_path

    def get_model_registry(self, args):
        model_path = self.get_model_path(args)
        if not model_path or args.dryrun:
            return ""

        parts = model_path.replace(args.store, "").split(os.sep)
        if len(parts) < 3:
            return ""
        return parts[2]

    def build_exec_args_bench(self, args, model_path):
        exec_model_path = MNT_FILE if args.container else model_path
        exec_args = ["llama-bench"]

        set_accel_env_vars()
        gpu_args = self.gpu_args(args=args)
        if gpu_args is not None:
            exec_args.extend(gpu_args)

        exec_args += ["-m", exec_model_path]

        return exec_args

    def build_exec_args_run(self, args, model_path, prompt):
        exec_model_path = model_path if not args.container else MNT_FILE

        # override prompt if not set to the local call
        if EMOJI and "LLAMA_PROMPT_PREFIX" not in os.environ:
            os.environ["LLAMA_PROMPT_PREFIX"] = "🦙 > "

        exec_args = ["llama-run", "-c", f"{args.context}", "--temp", f"{args.temp}"]
        exec_args += args.runtime_args

        if args.seed:
            exec_args += ["--seed", args.seed]

        if args.debug:
            exec_args += ["-v"]

        set_accel_env_vars()
        gpu_args = self.gpu_args(args=args, runner=True)
        if gpu_args is not None:
            exec_args.extend(gpu_args)

        if self.store is not None:
            ref_file = self.store.get_ref_file(self.tag)
            if ref_file.chat_template_name != "":
                exec_args.extend(["--chat-template-file", MNT_CHAT_TEMPLATE_FILE])

        exec_args.append(exec_model_path)
        if len(prompt) > 0:
            exec_args.append(prompt)

        return exec_args

    def execute_model(self, model_path, exec_args, args):
        try:
            if self.exec_model_in_container(model_path, exec_args, args):
                return
            if args.dryrun:
                dry_run(exec_args)
                return
            exec_cmd(exec_args, args.debug)
        except FileNotFoundError as e:
            if args.container:
                raise NotImplementedError(
                    file_not_found_in_container % {"cmd": exec_args[0], "error": str(e).strip("'")}
                )
            raise NotImplementedError(file_not_found % {"cmd": exec_args[0], "error": str(e).strip("'")})

    def validate_args(self, args):
        if args.container:
            return
        if args.privileged:
            raise KeyError(
                "--nocontainer and --privileged options conflict. The --privileged option requires a container."
            )
        if hasattr(args, "name") and args.name:
            if hasattr(args, "generate"):
                # Do not fail on serve if user specified --generate
                if args.generate:
                    return
            raise KeyError("--nocontainer and --name options conflict. The --name option requires a container.")

    def build_exec_args_serve(self, args, exec_model_path, chat_template_path=""):
        if args.runtime == "vllm":
            exec_args = [
                "--model",
                exec_model_path,
                "--port",
                args.port,
                "--max-sequence-length",
                f"{args.context}",
            ] + args.runtime_args
        else:
            exec_args = [
                "llama-server",
                "--port",
                args.port,
                "--model",
                exec_model_path,
                "--alias",
                self.model,
                "--ctx-size",
                f"{args.context}",
                "--temp",
                f"{args.temp}",
            ] + args.runtime_args
            if chat_template_path != "":
                exec_args.extend(["--chat-template-file", chat_template_path])

            if args.debug:
                exec_args.extend(["-v"])

        if args.seed:
            exec_args += ["--seed", args.seed]

        return exec_args

    def handle_runtime(self, args, exec_args, exec_model_path):
        set_accel_env_vars()
        if args.runtime == "vllm":
            exec_model_path = os.path.dirname(exec_model_path)
            # Left out "vllm", "serve" the image entrypoint already starts it
            exec_args = ["--port", args.port, "--model", MNT_FILE, "--max_model_len", "2048"]
        else:
            gpu_args = self.gpu_args(args=args)
            if gpu_args is not None:
                exec_args.extend(gpu_args)

            exec_args.extend(["--host", args.host])

        return exec_args

    def generate_container_config(self, model_path, chat_template_path, args, exec_args):
        self.image = accel_image(CONFIG, args)
        if args.generate == "quadlet":
            self.quadlet(model_path, chat_template_path, args, exec_args)
        elif args.generate == "kube":
            self.kube(model_path, chat_template_path, args, exec_args)
        elif args.generate == "quadlet/kube":
            self.quadlet_kube(model_path, chat_template_path, args, exec_args)
        else:
            return False

        return True

    def execute_command(self, model_path, exec_args, args):
        try:
            if self.exec_model_in_container(model_path, exec_args, args):
                return
            if args.dryrun:
                dry_run(exec_args)
                return
            exec_cmd(exec_args, debug=args.debug)
        except FileNotFoundError as e:
            if args.container:
                raise NotImplementedError(
                    file_not_found_in_container % {"cmd": exec_args[0], "error": str(e).strip("'")}
                )
            raise NotImplementedError(file_not_found % {"cmd": exec_args[0], "error": str(e).strip("'")})

    def serve(self, args, quiet=False):
        self.validate_args(args)
        args.port = compute_serving_port(args.port, args.debug, quiet)

        model_path = self.get_model_path(args)
        exec_model_path = MNT_FILE if args.container or args.generate else model_path

        chat_template_path = ""
        if self.store is not None:
            ref_file = self.store.get_ref_file(self.tag)
            if ref_file.chat_template_name != "":
                chat_template_path = (
                    MNT_CHAT_TEMPLATE_FILE
                    if args.container or args.generate
                    else self.store.get_snapshot_file_path(ref_file.hash, ref_file.chat_template_name)
                )

        exec_args = self.build_exec_args_serve(args, exec_model_path, chat_template_path)

        exec_args = self.handle_runtime(args, exec_args, exec_model_path)
        if self.generate_container_config(model_path, chat_template_path, args, exec_args):
            return

        # Add rag chatbot
        if hasattr(args, "rag") and args.rag:
            exec_args = [
                "bash",
                "-c",
                f"nohup {' '.join(exec_args)} &> /tmp/llama-server.log & rag_framework run /rag/vector.db",
            ]

        self.execute_command(model_path, exec_args, args)

    def quadlet(self, model, chat_template, args, exec_args):
        quadlet = Quadlet(model, chat_template, self.image, args, exec_args)
        quadlet.generate()

    def quadlet_kube(self, model, chat_template, args, exec_args):
        kube = Kube(model, chat_template, self.image, args, exec_args)
        kube.generate()
        quadlet = Quadlet(model, chat_template, self.image, args, exec_args)
        quadlet.kube()

    def kube(self, model, chat_template, args, exec_args):
        kube = Kube(model, chat_template, self.image, args, exec_args)
        kube.generate()

    def check_valid_model_path(self, relative_target_path, model_path):
        return os.path.exists(model_path) and os.readlink(model_path) == relative_target_path

    def inspect(self, args):
        model_name = self.filename
        model_path = self.get_model_path(args)
        model_registry = self.get_model_registry(args)

        if GGUFInfoParser.is_model_gguf(model_path):
            gguf_info: GGUFModelInfo = GGUFInfoParser.parse(model_name, model_registry, model_path, args)
            print(gguf_info.serialize(json=args.json, all=args.all))
            return

        print(ModelInfoBase(model_name, model_registry, model_path).serialize(json=args.json))


def dry_run(args):
    for arg in args:
        if not arg:
            continue
        if " " in arg:
            print('"%s"' % arg, end=" ")
        else:
            print("%s" % arg, end=" ")
    print()


def distinfo_volume():
    dist_info = "ramalama-%s.dist-info" % version()
    path = os.path.join(os.path.dirname(os.path.dirname(__file__)), dist_info)
    if not os.path.exists(path):
        return ""

    return f"-v{path}:/usr/share/ramalama/{dist_info}:ro"


def compute_ports() -> list:
    first_port = DEFAULT_PORT_RANGE[0]
    last_port = DEFAULT_PORT_RANGE[1]
    ports = list(range(first_port + 1, last_port + 1))
    random.shuffle(ports)
    # try always the first port before the randomized others
    return [first_port] + ports


def get_available_port_if_any(debug: bool) -> int:
    ports = compute_ports()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        chosen_port = 0
        for target_port in ports:
            if debug:
                print(f"Checking if {target_port} is available")
            try:
                s.bind(('localhost', target_port))
            except OSError:
                continue
            else:
                chosen_port = target_port
                break
        return chosen_port


def compute_serving_port(port: str, debug: bool, quiet=False) -> str:
    if not port:
        raise IOError("serving port can't be empty.")
    if port != int_tuple_as_str(DEFAULT_PORT_RANGE):
        # user specified a custom port, don't override the choice
        return port
    # otherwise compute a random serving port in the range
    target_port = get_available_port_if_any(debug)

    if target_port == 0:
        raise IOError("no available port could be detected. Please ensure you have enough free ports.")
    if not quiet:
        print(f"serving on port {target_port}")
    return str(target_port)



import os
import shutil
import subprocess
import tempfile

from ramalama.common import accel_image, get_accel_env_vars, run_cmd, set_accel_env_vars
from ramalama.config import CONFIG


class Rag:
    model = ""
    target = ""

    def __init__(self, target):
        self.target = target
        set_accel_env_vars()

    def build(self, source, target, args):
        print(f"Building {target}...")
        contextdir = os.path.dirname(source)
        src = os.path.basename(source)
        print(f"adding {src}...")
        cfile = f"""\
FROM scratch
COPY {src} /vector.db
"""
        containerfile = tempfile.NamedTemporaryFile(dir=source)
        # Open the file for writing.
        with open(containerfile.name, 'w') as c:
            c.write(cfile)
        if args.debug:
            print(f"\nContainerfile: {containerfile.name}\n{cfile}")
        imageid = (
            run_cmd(
                [
                    args.engine,
                    "build",
                    "--no-cache",
                    f"--network={args.network}",
                    "-q",
                    "-t",
                    target,
                    "-f",
                    containerfile.name,
                    contextdir,
                ],
                debug=args.debug,
            )
            .stdout.decode("utf-8")
            .strip()
        )
        return imageid

    def generate(self, args):
        args.image = accel_image(CONFIG, args)

        if not args.container:
            raise KeyError("rag command requires a container. Can not be run with --nocontainer option.")
        if not args.engine or args.engine == "":
            raise KeyError("rag command requires a container. Can not be run without a container engine.")

        # Default image with "-rag" append is used for building rag data.
        s = args.image.split(":")
        s[0] = s[0] + "-rag"
        rag_image = ":".join(s)

        exec_args = [args.engine, "run", "--rm"]
        if args.network:
            exec_args += ["--network", args.network]
        for path in args.PATH:
            if os.path.exists(path):
                fpath = os.path.realpath(path)
                exec_args += ["-v", f"{fpath}:/docs/{fpath}:ro,z"]
        tmpdir = "."
        if not os.access(tmpdir, os.W_OK):
            tmpdir = "/tmp"

        ragdb = tempfile.TemporaryDirectory(dir=tmpdir, prefix='RamaLama_rag_')
        dbdir = os.path.join(ragdb.name, "vectordb")
        os.mkdir(dbdir)
        exec_args += ["-v", f"{dbdir}:/output:z"]
        for k, v in get_accel_env_vars().items():
            # Special case for Cuda
            if k == "CUDA_VISIBLE_DEVICES":
                if os.path.basename(args.engine) == "docker":
                    exec_args += ["--gpus", "all"]
                else:
                    # newer Podman versions support --gpus=all, but < 5.0 do not
                    exec_args += ["--device", "nvidia.com/gpu=all"]

            exec_args += ["-e", f"{k}={v}"]

        exec_args += [rag_image]
        exec_args += ["doc2rag", "/output", "/docs/"]
        try:
            run_cmd(exec_args, debug=args.debug)
            print(self.build(dbdir, self.target, args))
        except subprocess.CalledProcessError as e:
            raise e
        finally:
            shutil.rmtree(ragdb.name, ignore_errors=True)



import argparse
from typing import Callable, Tuple, Union
from urllib.parse import urlparse

from ramalama.common import rm_until_substring
from ramalama.huggingface import Huggingface
from ramalama.model import MODEL_TYPES
from ramalama.model_store import GlobalModelStore, ModelStore
from ramalama.oci import OCI
from ramalama.ollama import Ollama
from ramalama.url import URL


class ModelFactory:

    def __init__(
        self,
        model: str,
        args: argparse,
        transport: str = "ollama",
        ignore_stderr: bool = False,
    ):
        self.model = model
        self.store_path = args.store
        self.use_model_store = args.use_model_store
        self.transport = transport
        self.engine = args.engine
        self.ignore_stderr = ignore_stderr
        self.container = args.container

        self.model_cls: type[Union[Huggingface, Ollama, OCI, URL]]
        self.create: Callable[[], Union[Huggingface, Ollama, OCI, URL]]
        self.model_cls, self.create = self.detect_model_model_type()

        self.pruned_model = self.prune_model_input()

    def detect_model_model_type(
        self,
    ) -> Tuple[type[Union[Huggingface, Ollama, OCI, URL]], Callable[[], Union[Huggingface, Ollama, OCI, URL]]]:
        if self.model.startswith("huggingface://") or self.model.startswith("hf://") or self.model.startswith("hf.co/"):
            return Huggingface, self.create_huggingface
        if self.model.startswith("ollama://") or "ollama.com/library/" in self.model:
            return Ollama, self.create_ollama
        if self.model.startswith("oci://") or self.model.startswith("docker://"):
            return OCI, self.create_oci
        if self.model.startswith("http://") or self.model.startswith("https://") or self.model.startswith("file://"):
            return URL, self.create_url

        if self.transport == "huggingface":
            return Huggingface, self.create_huggingface
        if self.transport == "ollama":
            return Ollama, self.create_ollama
        if self.transport == "oci":
            return OCI, self.create_oci

        raise KeyError(f'transport "{self.transport}" not supported. Must be oci, huggingface, or ollama.')

    def prune_model_input(self) -> str:
        # remove protocol from model input
        pruned_model_input = rm_until_substring(self.model, "://")

        if self.model_cls == Huggingface:
            pruned_model_input = rm_until_substring(pruned_model_input, "hf.co/")
        elif self.model_cls == Ollama:
            pruned_model_input = rm_until_substring(pruned_model_input, "ollama.com/library/")

        return pruned_model_input

    def validate_oci_model_input(self):
        if self.model.startswith("oci://") or self.model.startswith("docker://"):
            return

        for t in MODEL_TYPES:
            if self.model.startswith(t + "://"):
                raise ValueError(f"{self.model} invalid: Only OCI Model types supported")

    def set_optional_model_store(self, model: Union[Huggingface, Ollama, OCI, URL]):
        if self.use_model_store:
            name, _, orga = model.extract_model_identifiers()
            model.store = ModelStore(GlobalModelStore(self.store_path), name, type(model).__name__.lower(), orga)

    def create_huggingface(self) -> Huggingface:
        model = Huggingface(self.pruned_model)
        self.set_optional_model_store(model)
        return model

    def create_ollama(self) -> Ollama:
        model = Ollama(self.pruned_model)
        self.set_optional_model_store(model)
        return model

    def create_oci(self) -> OCI:
        if not self.container:
            raise ValueError("OCI containers cannot be used with the --nocontainer option.")

        self.validate_oci_model_input()
        model = OCI(self.pruned_model, self.engine, self.ignore_stderr)
        self.set_optional_model_store(model)
        return model

    def create_url(self) -> URL:
        model = URL(self.pruned_model, urlparse(self.model).scheme)
        self.set_optional_model_store(model)
        return model



import json
import os
import pathlib
import shutil
import tempfile
import urllib.request

from ramalama.common import available, download_file, exec_cmd, generate_sha256, perror, run_cmd, verify_checksum
from ramalama.model import Model
from ramalama.model_store import SnapshotFile, SnapshotFileType

missing_huggingface = """
Optional: Huggingface models require the huggingface-cli module.
These modules can be installed via PyPi tools like pip, pip3, pipx, or via
distribution package managers like dnf or apt. Example:
pip install huggingface_hub
"""


def is_huggingface_cli_available():
    """Check if huggingface-cli is available on the system."""
    return available("huggingface-cli")


def fetch_checksum_from_api(organization, file):
    """Fetch the SHA-256 checksum from the model's metadata API for a given file."""
    checksum_api_url = f"{HuggingfaceRepository.REGISTRY_URL}/{organization}/raw/main/{file}"
    try:
        with urllib.request.urlopen(checksum_api_url) as response:
            data = response.read().decode()
        # Extract the SHA-256 checksum from the `oid sha256` line
        for line in data.splitlines():
            if line.startswith("oid sha256:"):
                return line.split(":", 1)[1].strip()
        raise ValueError("SHA-256 checksum not found in the API response.")
    except urllib.error.HTTPError as e:
        raise KeyError(f"failed to pull {checksum_api_url}: " + str(e).strip("'"))
    except urllib.error.URLError as e:
        raise KeyError(f"failed to pull {checksum_api_url}: " + str(e).strip("'"))


class HuggingfaceCLIFile(SnapshotFile):

    def __init__(
        self, url, header, hash, name, type, should_show_progress=False, should_verify_checksum=False, required=True
    ):
        super().__init__(url, header, hash, name, type, should_show_progress, should_verify_checksum, required)

    def download(self, blob_file_path, snapshot_dir):
        # moving from the cached temp directory to blob directory
        shutil.move(self.url, blob_file_path)
        return os.path.relpath(blob_file_path, start=snapshot_dir)


class HuggingfaceRepository:

    REGISTRY_URL = "https://huggingface.co"

    FILE_NAME_CONFIG = "config.json"
    FILE_NAME_GENERATION_CONFIG = "generation_config.json"
    FILE_NAME_TOKENIZER_CONFIG = "tokenizer_config.json"

    def __init__(self, name: str, organization: str):
        self.name = name
        self.organization = organization

        self.blob_url = f"{HuggingfaceRepository.REGISTRY_URL}/{self.organization}/resolve/main"
        self.headers = {}

    def get_file_list(self, cached_files: list[str], snapshot_hash: str) -> list[SnapshotFile]:
        files = []
        if self.name not in cached_files:
            files.append(self.model_file(snapshot_hash))
        if HuggingfaceRepository.FILE_NAME_CONFIG not in cached_files:
            files.append(self.config_file())
        if HuggingfaceRepository.FILE_NAME_GENERATION_CONFIG not in cached_files:
            files.append(self.generation_config_file())
        if HuggingfaceRepository.FILE_NAME_TOKENIZER_CONFIG not in cached_files:
            files.append(self.tokenizer_config_file())

        return files

    def model_file(self, snapshot_hash: str) -> SnapshotFile:
        return SnapshotFile(
            url=f"{self.blob_url}/{self.name}",
            header=self.headers,
            hash=snapshot_hash,
            type=SnapshotFileType.Model,
            name=self.name,
            should_show_progress=True,
            should_verify_checksum=True,
        )

    def config_file(self) -> SnapshotFile:
        return SnapshotFile(
            url=f"{self.blob_url}/{HuggingfaceRepository.FILE_NAME_CONFIG}",
            header=self.headers,
            hash=generate_sha256(HuggingfaceRepository.FILE_NAME_CONFIG),
            type=SnapshotFileType.Other,
            name=HuggingfaceRepository.FILE_NAME_CONFIG,
            required=False,
        )

    def generation_config_file(self) -> SnapshotFile:
        return SnapshotFile(
            url=f"{self.blob_url}/{HuggingfaceRepository.FILE_NAME_GENERATION_CONFIG}",
            header=self.headers,
            hash=generate_sha256(HuggingfaceRepository.FILE_NAME_GENERATION_CONFIG),
            type=SnapshotFileType.Other,
            name=HuggingfaceRepository.FILE_NAME_GENERATION_CONFIG,
            required=False,
        )

    def tokenizer_config_file(self) -> SnapshotFile:
        return SnapshotFile(
            url=f"{self.blob_url}/{HuggingfaceRepository.FILE_NAME_TOKENIZER_CONFIG}",
            header=self.headers,
            hash=generate_sha256(HuggingfaceRepository.FILE_NAME_TOKENIZER_CONFIG),
            type=SnapshotFileType.Other,
            name=HuggingfaceRepository.FILE_NAME_TOKENIZER_CONFIG,
            required=False,
        )


def get_repo_info(repo_name):
    # Docs on API call:
    # https://huggingface.co/docs/hub/en/api#get-apimodelsrepoid-or-apimodelsrepoidrevisionrevision
    repo_info_url = f"https://huggingface.co/api/models/{repo_name}"
    with urllib.request.urlopen(repo_info_url) as response:
        if response.getcode() == 200:
            repo_info = response.read().decode('utf-8')
            return json.loads(repo_info)
        else:
            perror("Huggingface repo information pull failed")
            raise KeyError(f"Response error code from repo info pull: {response.getcode()}")
    return None


def handle_repo_info(repo_name, repo_info, runtime):
    if "safetensors" in repo_info and runtime == "llama.cpp":
        print(
            "\nllama.cpp does not support running safetensor models, "
            "please use a/convert to the GGUF format using:\n"
            f"- https://huggingface.co/models?other=base_model:quantized:{repo_name} \n"
            "- https://huggingface.co/spaces/ggml-org/gguf-my-repo"
        )
    if "gguf" in repo_info:
        print("There are GGUF files to choose from in this repo, use one of the following commands to run one:\n")
    for sibling in repo_info.get("siblings", []):
        if sibling["rfilename"].endswith('.gguf'):
            file = sibling["rfilename"]
            print(f"- ramalama run hf://{repo_name}/{file}")
    print("\n")


class Huggingface(Model):
    def __init__(self, model):
        super().__init__(model)

        self.type = "huggingface"
        self.hf_cli_available = is_huggingface_cli_available()

    def login(self, args):
        if not self.hf_cli_available:
            raise NotImplementedError(missing_huggingface)
        conman_args = ["huggingface-cli", "login"]
        if args.token:
            conman_args.extend(["--token", args.token])
        self.exec(conman_args, args)

    def logout(self, args):
        if not self.hf_cli_available:
            raise NotImplementedError(missing_huggingface)
        conman_args = ["huggingface-cli", "logout"]
        if args.token:
            conman_args.extend(["--token", args.token])
        self.exec(conman_args, args)

    def pull(self, args):
        if self.store is not None:
            return self._pull_with_model_store()

        model_path = self.model_path(args)
        directory_path = os.path.join(args.store, "repos", "huggingface", self.directory, self.filename)
        os.makedirs(directory_path, exist_ok=True)

        symlink_dir = os.path.dirname(model_path)
        os.makedirs(symlink_dir, exist_ok=True)

        try:
            # Check if huggingface repo instead of file
            if self.directory.count("/") == 0:
                repo_name = self.directory + "/" + self.filename
                repo_info = get_repo_info(repo_name)
                handle_repo_info(repo_name, repo_info, args.runtime)

            return self.url_pull(args, model_path, directory_path)
        except (urllib.error.HTTPError, urllib.error.URLError, KeyError) as e:
            if self.hf_cli_available:
                return self.hf_pull(args, model_path, directory_path)
            perror("URL pull failed and huggingface-cli not available")
            raise KeyError(f"Failed to pull model: {str(e)}")

    def _fetch_snapshot_path(self, cache_dir, namespace, repo):
        cache_path = os.path.join(cache_dir, f'models--{namespace}--{repo}')
        main_ref_path = os.path.join(cache_path, 'refs', 'main')
        if not (os.path.exists(cache_path) and os.path.exists(main_ref_path)):
            return None, None
        with open(main_ref_path, 'r') as file:
            snapshot = file.read().strip()
        snapshot_path = os.path.join(cache_path, 'snapshots', snapshot)
        return snapshot_path, cache_path

    def in_existing_cache(self, args, target_path, sha256_checksum):
        if not self.hf_cli_available:
            return False

        default_hf_caches = [os.path.join(os.environ['HOME'], '.cache/huggingface/hub')]
        namespace, repo = os.path.split(str(self.directory))

        for cache_dir in default_hf_caches:
            snapshot_path, cache_path = self._fetch_snapshot_path(cache_dir, namespace, repo)
            if not snapshot_path or not os.path.exists(snapshot_path):
                continue

            file_path = os.path.join(snapshot_path, self.filename)
            if not os.path.exists(file_path):
                continue

            blob_path = pathlib.Path(file_path).resolve()
            if not os.path.exists(blob_path):
                continue

            blob_file = os.path.relpath(blob_path, start=os.path.join(cache_path, 'blobs'))
            if str(blob_file) != str(sha256_checksum):
                continue

            os.symlink(blob_path, target_path)
            return True
        return False

    def hf_pull(self, args, model_path, directory_path):
        conman_args = ["huggingface-cli", "download", "--local-dir", directory_path, self.model]
        run_cmd(conman_args, debug=args.debug)

        relative_target_path = os.path.relpath(directory_path, start=os.path.dirname(model_path))
        pathlib.Path(model_path).unlink(missing_ok=True)
        os.symlink(relative_target_path, model_path)
        return model_path

    def url_pull(self, args, model_path, directory_path):
        # Fetch the SHA-256 checksum from the API
        sha256_checksum = fetch_checksum_from_api(self.directory, self.filename)

        target_path = os.path.join(directory_path, f"sha256:{sha256_checksum}")

        if not os.path.exists(target_path):
            self.in_existing_cache(args, target_path, sha256_checksum)

        if os.path.exists(target_path) and verify_checksum(target_path):
            relative_target_path = os.path.relpath(target_path, start=os.path.dirname(model_path))
            if not self.check_valid_model_path(relative_target_path, model_path):
                pathlib.Path(model_path).unlink(missing_ok=True)
                os.symlink(relative_target_path, model_path)
            return model_path

        # Download the model file to the target path
        url = f"https://huggingface.co/{self.directory}/resolve/main/{self.filename}"
        download_file(url, target_path, headers={}, show_progress=True)
        if not verify_checksum(target_path):
            print(f"Checksum mismatch for {target_path}, retrying download...")
            os.remove(target_path)
            download_file(url, target_path, headers={}, show_progress=True)
            if not verify_checksum(target_path):
                raise ValueError(f"Checksum verification failed for {target_path}")

        relative_target_path = os.path.relpath(target_path, start=os.path.dirname(model_path))
        if self.check_valid_model_path(relative_target_path, model_path):
            # Symlink is already correct, no need to update it
            return model_path

        pathlib.Path(model_path).unlink(missing_ok=True)
        os.symlink(relative_target_path, model_path)
        return model_path

    def push(self, source, args):
        if not self.hf_cli_available:
            raise NotImplementedError(missing_huggingface)
        proc = run_cmd(
            [
                "huggingface-cli",
                "upload",
                "--repo-type",
                "model",
                self.directory,
                self.filename,
                "--cache-dir",
                os.path.join(args.store, "repos", "huggingface", ".cache"),
                "--local-dir",
                os.path.join(args.store, "repos", "huggingface", self.directory),
            ],
            debug=args.debug,
        )
        return proc.stdout.decode("utf-8")

    def exec(self, cmd_args, args):
        try:
            exec_cmd(cmd_args, debug=args.debug)
        except FileNotFoundError as e:
            print(f"{str(e).strip()}\n{missing_huggingface}")

    def _collect_cli_files(self, tempdir: str) -> tuple[str, list[HuggingfaceCLIFile]]:
        cache_dir = os.path.join(tempdir, ".cache", "huggingface", "download")
        files: list[HuggingfaceCLIFile] = []
        snapshot_hash = ""
        for entry in os.listdir(tempdir):
            entry_path = os.path.join(tempdir, entry)
            if os.path.isdir(entry_path) or entry == ".gitattributes":
                continue
            sha256 = ""
            with open(os.path.join(cache_dir, f"{entry}.metadata")) as metafile:
                metafile.readline()
                sha256 = f"sha256:{metafile.readline().strip()}"
            if sha256 == "sha256:":
                continue
            if entry.lower() == "readme.md":
                snapshot_hash = sha256
                continue
            files.append(
                HuggingfaceCLIFile(
                    url=entry_path,
                    header={},
                    hash=sha256,
                    type=SnapshotFileType.Other,
                    name=entry,
                )
            )

        return snapshot_hash, files

    def _pull_with_model_store(self, debug: bool = False):
        name, tag, organization = self.extract_model_identifiers()
        hash, cached_files, all = self.store.get_cached_files(tag)
        if all:
            return self.store.get_snapshot_file_path(hash, name)

        try:
            # Fetch the SHA-256 checksum of model from the API and use as snapshot hash
            snapshot_hash = f"sha256:{fetch_checksum_from_api(self.organization, self.name)}"

            hf_repo = HuggingfaceRepository(name, organization)
            files = hf_repo.get_file_list(cached_files, snapshot_hash)
            self.store.new_snapshot(tag, snapshot_hash, files)
        except Exception as e:
            if not self.hf_cli_available:
                perror("URL pull failed and huggingface-cli not available")
                raise KeyError(f"Failed to pull model: {str(e)}")

            # Cleanup previously created snapshot
            self.store.remove_snapshot(tag)

            # Create temporary directory for downloading via huggingface-cli
            with tempfile.TemporaryDirectory() as tempdir:
                model = f"{organization}/{name}"
                conman_args = ["huggingface-cli", "download", "--local-dir", tempdir, model]
                run_cmd(conman_args, debug=debug)

                snapshot_hash, files = self._collect_cli_files(tempdir)
                self.store.new_snapshot(tag, snapshot_hash, files)

        return self.store.get_snapshot_file_path(snapshot_hash, self.store.model_name)



import os
import shutil

from ramalama.common import download_file, generate_sha256
from ramalama.huggingface import HuggingfaceRepository
from ramalama.model import Model
from ramalama.model_store import SnapshotFile, SnapshotFileType


class LocalModelFile(SnapshotFile):

    def __init__(
        self, url, header, hash, name, should_show_progress=False, should_verify_checksum=False, required=True
    ):
        super().__init__(
            url, header, hash, name, SnapshotFileType.Model, should_show_progress, should_verify_checksum, required
        )

    def download(self, blob_file_path, snapshot_dir):
        if not os.path.exists(self.url):
            raise FileNotFoundError(f"No such file: '{self.url}'")
        # moving from the local location to blob directory so the model store "owns" the data
        shutil.copy(self.url, blob_file_path)
        return os.path.relpath(blob_file_path, start=snapshot_dir)


class URL(Model):
    def __init__(self, model, scheme):
        super().__init__(model)

        self.type = scheme
        split = self.model.rsplit("/", 1)
        self.directory = split[0].removeprefix("/") if len(split) > 1 else ""

    def extract_model_identifiers(self):
        model_name, model_tag, model_organization = super().extract_model_identifiers()

        parts = model_organization.split("/")
        if len(parts) > 2 and parts[-2] == "blob":
            model_organization = "/".join(parts[:-2])
            model_tag = parts[-1]

        # handling huggingface specific URLs for more precise identifiers
        if len(parts) > 2 and HuggingfaceRepository.REGISTRY_URL.endswith(parts[0]) and parts[-2] == "resolve":
            model_organization = "/".join(parts[:-2])
            model_tag = parts[-1]

        return model_name, model_tag, model_organization

    def pull(self, args):
        if self.store is not None:
            return self._pull_with_model_store()

        model_path = self.model_path(args)
        directory_path = os.path.join(args.store, "repos", self.type, self.directory, self.filename)
        os.makedirs(directory_path, exist_ok=True)

        symlink_dir = os.path.dirname(model_path)
        os.makedirs(symlink_dir, exist_ok=True)

        target_path = os.path.join(directory_path, self.filename)

        if self.type == "file":
            if not os.path.exists(self.model):
                raise FileNotFoundError(f"{self.model} no such file")
            os.symlink(self.model, os.path.join(symlink_dir, self.filename))
            os.symlink(self.model, target_path)
        else:
            show_progress = not args.quiet
            url = self.type + "://" + self.model
            # Download the model file to the target path
            download_file(url, target_path, headers={}, show_progress=show_progress)
            relative_target_path = os.path.relpath(target_path, start=os.path.dirname(model_path))
            if self.check_valid_model_path(relative_target_path, model_path):
                # Symlink is already correct, no need to update it
                return model_path
            os.symlink(relative_target_path, model_path)

        return model_path

    def _pull_with_model_store(self):
        name, tag, _ = self.extract_model_identifiers()
        hash, _, all = self.store.get_cached_files(tag)
        if all:
            return self.store.get_snapshot_file_path(hash, name)

        files: list[SnapshotFile] = []
        snapshot_hash = generate_sha256(name)
        if self.type == "file":
            files.append(
                LocalModelFile(
                    url=self.model,  # model contains the full path here
                    header={},
                    hash=snapshot_hash,
                    name=name,
                    required=True,
                )
            )
        else:
            files.append(
                SnapshotFile(
                    url=f"{self.type}://{self.model}",
                    header={},
                    hash=snapshot_hash,
                    type=SnapshotFileType.Model,
                    name=name,
                    should_show_progress=True,
                    required=True,
                )
            )

        self.store.new_snapshot(tag, snapshot_hash, files)

        return self.store.get_snapshot_file_path(snapshot_hash, name)



"""ramalama common module."""

import glob
import hashlib
import json
import logging
import os
import random
import shutil
import string
import subprocess
import sys
import time
import urllib.error
import urllib.request
from typing import List

import ramalama.console as console
from ramalama.http_client import HttpClient
from ramalama.version import version

logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(levelname)s - %(message)s")

MNT_DIR = "/mnt/models"
MNT_FILE = f"{MNT_DIR}/model.file"
MNT_CHAT_TEMPLATE_FILE = f"{MNT_DIR}/chat_template.file"

RAG_DIR = "/rag"
RAG_CONTENT = f"{MNT_DIR}/vector.db"

HTTP_NOT_FOUND = 404
HTTP_RANGE_NOT_SATISFIABLE = 416  # "Range Not Satisfiable" error (file already downloaded)

DEFAULT_IMAGE = "quay.io/ramalama/ramalama"


_engine = -1  # -1 means cached variable not set yet
_nvidia = -1  # -1 means cached variable not set yet
podman_machine_accel = False


def get_engine():
    engine = os.getenv("RAMALAMA_CONTAINER_ENGINE")
    if engine is not None:
        if os.path.basename(engine) == "podman" and sys.platform == "darwin":
            # apple_vm triggers setting global variable podman_machine_accel side effect
            apple_vm(engine)
        return engine

    if available("podman") and (sys.platform != "darwin" or apple_vm("podman")):
        return "podman"

    if available("docker") and sys.platform != "darwin":
        return "docker"

    return None


def container_manager():
    global _engine
    if _engine != -1:
        return _engine

    _engine = get_engine()

    return _engine


def confirm_no_gpu(name, provider):
    while True:
        user_input = (
            input(
                f"Warning! Your VM {name} is using {provider}, which does not support GPU. "
                "Only the provider libkrun has GPU support. "
                "Do you want to proceed without GPU? (yes/no): "
            )
            .strip()
            .lower()
        )
        if user_input in ["yes", "y"]:
            return True
        if user_input in ["no", "n"]:
            return False
        print("Invalid input. Please enter 'yes' or 'no'.")


def handle_provider(machine):
    global podman_machine_accel
    name = machine.get("Name")
    provider = machine.get("VMType")
    running = machine.get("Running")
    if running:
        if provider == "applehv":
            return confirm_no_gpu(name, provider)
        if "krun" in provider:
            podman_machine_accel = True
            return True

    return None


def apple_vm(engine):
    podman_machine_list = [engine, "machine", "list", "--format", "json", "--all-providers"]
    try:
        machines_json = run_cmd(podman_machine_list, ignore_stderr=True).stdout.decode("utf-8").strip()
        machines = json.loads(machines_json)
        for machine in machines:
            result = handle_provider(machine)
            if result is not None:
                return result
    except subprocess.CalledProcessError:
        pass
    return False


def perror(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def available(cmd):
    return shutil.which(cmd) is not None


def exec_cmd(args, debug=False):
    if debug:
        perror("exec_cmd: ", *args)

    try:
        return os.execvp(args[0], args)
    except Exception:
        perror(f"os.execvp({args[0]}, {args})")
        raise


def run_cmd(args, cwd=None, stdout=subprocess.PIPE, ignore_stderr=False, ignore_all=False, debug=False):
    """
    Run the given command arguments.

    Args:
    args: command line arguments to execute in a subprocess
    cwd: optional working directory to run the command from
    stdout: standard output configuration
    ignore_stderr: if True, ignore standard error
    ignore_all: if True, ignore both standard output and standard error
    debug: if True, print debug information
    """
    if debug:
        perror("run_cmd: ", *args)
        perror(f"Working directory: {cwd}")
        perror(f"Ignore stderr: {ignore_stderr}")
        perror(f"Ignore all: {ignore_all}")

    serr = None
    if ignore_all or ignore_stderr:
        serr = subprocess.DEVNULL

    sout = subprocess.PIPE
    if ignore_all:
        sout = subprocess.DEVNULL

    result = subprocess.run(args, check=True, cwd=cwd, stdout=sout, stderr=serr)
    if debug:
        print("Command finished with return code:", result.returncode)

    return result


def find_working_directory():
    return os.path.dirname(__file__)


def generate_sha256(to_hash: str) -> str:
    """
    Generates a sha256 for a string.

    Args:
    to_hash (str): The string to generate the sha256 hash for.

    Returns:
    str: Hex digest of the input appended to the prefix sha256:
    """
    h = hashlib.new("sha256")
    h.update(to_hash.encode("utf-8"))
    return f"sha256:{h.hexdigest()}"


def verify_checksum(filename):
    """
    Verifies if the SHA-256 checksum of a file matches the checksum provided in
    the filename.

    Args:
    filename (str): The filename containing the checksum prefix
                    (e.g., "sha256:<checksum>")

    Returns:
    bool: True if the checksum matches, False otherwise.
    """

    if not os.path.exists(filename):
        return False

    # Check if the filename starts with "sha256:" or "sha256-" and extract the checksum from filename
    expected_checksum = ""
    fn_base = os.path.basename(filename)
    if fn_base.startswith("sha256:"):
        expected_checksum = fn_base.split(":")[1]
    elif fn_base.startswith("sha256-"):
        expected_checksum = fn_base.split("-")[1]
    else:
        raise ValueError(f"filename has to start with 'sha256:' or 'sha256-': {fn_base}")

    if len(expected_checksum) != 64:
        raise ValueError("invalid checksum length in filename")

    # Calculate the SHA-256 checksum of the file contents
    sha256_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    # Compare the checksums
    return sha256_hash.hexdigest() == expected_checksum


# default_image function should figure out which GPU the system uses t
# then running appropriate container image.
def default_image():
    image = os.getenv("RAMALAMA_IMAGE")
    if image:
        return image

    return DEFAULT_IMAGE


def genname():
    return "ramalama_" + "".join(random.choices(string.ascii_letters + string.digits, k=10))


def download_file(url, dest_path, headers=None, show_progress=True):
    """
    Downloads a file from a given URL to a specified destination path.

    Args:
        url (str): The URL to download from.
        dest_path (str): The path to save the downloaded file.
        headers (dict): Optional headers to include in the request.
        show_progress (bool): Whether to show a progress bar during download.

    Raises:
        RuntimeError: If the download fails after multiple attempts.
    """
    headers = headers or {}

    # If not running in a TTY, disable progress to prevent CI pollution
    if not sys.stdout.isatty():
        show_progress = False

    http_client = HttpClient()
    max_retries = 5  # Stop after 5 failures
    retries = 0

    while retries < max_retries:
        try:
            # Initialize HTTP client for the request
            http_client.init(url=url, headers=headers, output_file=dest_path, show_progress=show_progress)
            return  # Exit function if successful

        except urllib.error.HTTPError as e:
            if e.code in [HTTP_RANGE_NOT_SATISFIABLE, HTTP_NOT_FOUND]:
                raise e
            retries += 1

        except urllib.error.URLError as e:
            console.error(f"Network Error: {e.reason}")
            retries += 1

        except TimeoutError:
            retries += 1
            console.warning(f"TimeoutError: The server took too long to respond. Retrying {retries}/{max_retries}...")

        except RuntimeError as e:  # Catch network-related errors from HttpClient
            retries += 1
            console.warning(f"{e}. Retrying {retries}/{max_retries}...")

        except IOError as e:
            retries += 1
            console.warning(f"I/O Error: {e}. Retrying {retries}/{max_retries}...")

        except Exception as e:
            console.error(f"Unexpected error: {str(e)}")
            raise e

        if retries >= max_retries:
            error_message = (
                "\nDownload failed after multiple attempts.\n"
                "Possible causes:\n"
                "- Internet connection issue\n"
                "- Server is down or unresponsive\n"
                "- Firewall or proxy blocking the request\n"
            )
            raise Exception(error_message)

        time.sleep(2**retries * 0.1)  # Exponential backoff (0.1s, 0.2s, 0.4s...)


def engine_version(engine):
    # Create manifest list for target with imageid
    cmd_args = [engine, "version", "--format", "{{ .Client.Version }}"]
    return run_cmd(cmd_args).stdout.decode("utf-8").strip()


def resolve_cdi(spec_dirs: List[str]):
    """Loads all CDI specs from the given directories."""
    for spec_dir in spec_dirs:
        for root, _, files in os.walk(spec_dir):
            for file in files:
                if file.endswith('.json') or file.endswith('.yaml'):
                    if load_spec(os.path.join(root, file)):
                        return True

    return False


def yaml_safe_load(stream) -> dict:
    data = {}
    for line in stream:
        if ':' in line:
            key, value = line.split(':', 1)
            data[key.strip()] = value.strip()

    return data


def load_spec(path: str):
    """Loads a single CDI spec file."""
    with open(path, 'r') as f:
        spec = json.load(f) if path.endswith('.json') else yaml_safe_load(f)

    return spec.get('kind')


def check_asahi():
    if os.path.exists('/proc/device-tree/compatible'):
        try:
            with open('/proc/device-tree/compatible', 'rb') as f:
                content = f.read().split(b"\0")
                if b"apple,arm-platform" in content:
                    os.environ["ASAHI_VISIBLE_DEVICES"] = "1"
                    return "asahi"
        except OSError:
            pass

    return None


def check_nvidia():
    global _nvidia
    if _nvidia != -1:
        return _nvidia

    try:
        command = ['nvidia-smi']
        run_cmd(command).stdout.decode("utf-8")

        # ensure at least one CDI device resolves
        if resolve_cdi(['/etc/cdi', '/var/run/cdi']):
            os.environ["CUDA_VISIBLE_DEVICES"] = "0"
            _nvidia = "cuda"
            return _nvidia

    except Exception:
        _nvidia = ""

    return _nvidia


def check_ascend():
    try:
        command = ['npu-smi', 'info']
        run_cmd(command).stdout.decode("utf-8")
        os.environ["ASCEND_VISIBLE_DEVICES"] = "0"
        return "cann"
    except Exception:
        pass

    return None


def check_rocm_amd():
    gpu_num = 0
    gpu_bytes = 0
    for i, fp in enumerate(sorted(glob.glob('/sys/bus/pci/devices/*/mem_info_vram_total'))):
        with open(fp, 'r') as file:
            content = int(file.read())
            if content > 1073741824 and content > gpu_bytes:
                gpu_bytes = content
                gpu_num = i

    if gpu_bytes:
        os.environ["HIP_VISIBLE_DEVICES"] = str(gpu_num)
        return "hip"

    return None


def check_intel():
    igpu_num = 0
    # Device IDs for select Intel GPUs.  See: https://dgpu-docs.intel.com/devices/hardware-table.html
    intel_gpus = (
        b"0xe20b",
        b"0xe20c",
        b"0x7d51",
        b"0x7dd5",
        b"0x7d55",
    )
    # Check to see if any of the device ids in intel_gpus are in the device id of the i915 driver
    for fp in sorted(glob.glob('/sys/bus/pci/drivers/i915/*/device')):
        with open(fp, 'rb') as file:
            content = file.read()
            for gpu_id in intel_gpus:
                if gpu_id in content:
                    igpu_num += 1
    if igpu_num:
        os.environ["INTEL_VISIBLE_DEVICES"] = str(igpu_num)
        return "intel"

    return None


def get_accel():
    if gpu_type := check_asahi():
        return gpu_type

    if gpu_type := check_nvidia():
        return gpu_type

    if gpu_type := check_ascend():
        return gpu_type

    if gpu_type := check_rocm_amd():
        return gpu_type

    if gpu_type := check_intel():
        return gpu_type

    return "none"


def set_accel_env_vars():
    if get_accel_env_vars():
        return

    get_accel()


def get_accel_env_vars():
    gpu_vars = (
        "ASAHI_VISIBLE_DEVICES",
        "ASCEND_VISIBLE_DEVICES",
        "CUDA_VISIBLE_DEVICES",
        "CUDA_LAUNCH_BLOCKING",
        "HIP_VISIBLE_DEVICES",
        "HSA_VISIBLE_DEVICES",
        "HSA_OVERRIDE_GFX_VERSION",
        "INTEL_VISIBLE_DEVICES",
    )
    env_vars = {k: v for k, v in os.environ.items() for gpu_var in gpu_vars if k == gpu_var}

    return env_vars


def rm_until_substring(input, substring):
    pos = input.find(substring)
    if pos == -1:
        return input

    # Create a new string starting after the found substring
    return ''.join(input[i] for i in range(pos + len(substring), len(input)))


def accel_image(config, args):
    if args.image != DEFAULT_IMAGE:
        return args.image

    env_vars = get_accel_env_vars()

    if not env_vars:
        gpu_type = None
    else:
        gpu_type, _ = next(iter(env_vars.items()))

    if args.runtime == "vllm":
        if "HIP_VISIBLE_DEVICES" in os.environ:
            return "docker.io/rocm/vllm-dev:main"

        return "docker.io/vllm/vllm-openai"

    split = version().split(".")
    vers = ".".join(split[:2])
    conman = config['engine']
    images = config['images']
    image = images.get(gpu_type, args.image)
    if attempt_to_use_versioned(conman, image, vers, args.debug):
        return f"{image}:{vers}"

    return f"{image}:latest"


def attempt_to_use_versioned(conman, image, vers, debug):
    try:
        # check if versioned image exists locally
        if run_cmd([conman, "inspect", f"{image}:{vers}"], ignore_all=True, debug=debug):
            return True

    except Exception:
        pass

    try:
        # attempt to pull the versioned image
        run_cmd([conman, "pull", f"{image}:{vers}"], ignore_stderr=True, debug=debug)
        return True

    except Exception:
        return False
